/*
 *  Copyright (C) 2002-2005  The DOSBox Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* $Id: directserial_win32.cpp,v 1.5 2005/02/10 10:21:08 qbix79 Exp $ */

#include "dosbox.h"

#if C_DIRECTSERIAL

/* Windows version */
#if defined (WIN32)

//#include <stdarg.h>
#include <string.h>
#include <ctype.h>

#include "setup.h"
#include "serialport.h"
#include "ds.h"

// Win32 related headers
#include <windows.h>

/* This is a serial passthrough class.  Its amazingly simple to */
/* write now that the serial ports themselves were abstracted out */


//class CDirectSerial : public CSerial {
//public:
//	HANDLE hCom;
///	DCB dcb;
//	BOOL fSuccess;

	CDirectSerial::CDirectSerial(	
					Bit16u baseAddr,
					Bit8u initIrq,
					Bit32u initBps,
					Bit8u bytesize,
					const char *parity,
					Bit8u stopbits,
					const char * realPort
		) : CSerial(baseAddr,
					initIrq,
					initBps,
					bytesize,
					parity,
					stopbits)
	{
		
		//LOG_MSG("Opening Windows serial port");
		hCom = CreateFile(
					realPort,
					GENERIC_READ | GENERIC_WRITE,
                    0,					// must be opened with exclusive-access
                    NULL,				// no security attributes
                    OPEN_EXISTING,		// must use OPEN_EXISTING
                    0,					// non overlapped I/O
                    NULL				// hTemplate must be NULL for comm devices
                    );

		if (hCom == INVALID_HANDLE_VALUE) 
		{
			int error = GetLastError();
			LOG_MSG("Serial port \"%s\" could not be opened.",realPort);
			if(error==2)
			{
				LOG_MSG("The specified port does not exist.");
			}
			else if(error==5)
			{
				LOG_MSG("The specified port is already in use.");
			}
			else 
			{
				LOG_MSG("Windows error %d occurred.",error);
			}
			
			hCom = 0;
			return;
		}

		fSuccess = GetCommState(hCom, &dcb);

		if (!fSuccess) 
		{
			// Handle the error.
			LOG_MSG("GetCommState failed with error %d.\n", GetLastError());
			hCom = 0;
			return;
		}

		// Configure timeouts to effectively use polling
		COMMTIMEOUTS ct;
		ct.ReadIntervalTimeout = MAXDWORD;
		ct.ReadTotalTimeoutConstant = 0;
		ct.ReadTotalTimeoutMultiplier = 0;
		ct.WriteTotalTimeoutConstant = 0;
		ct.WriteTotalTimeoutMultiplier = 0;
		SetCommTimeouts(hCom, &ct);

		CSerial::Init_Registers(initBps,bytesize,parity,stopbits);

	}

	CDirectSerial::~CDirectSerial()
	{
		if(hCom != INVALID_HANDLE_VALUE) CloseHandle(hCom);
	}

	bool CDirectSerial::CanRecv(void) { return true; }
	bool CDirectSerial::CanSend(void) { return true; }

	void CDirectSerial::RXBufferEmpty()
	{	// no external buffer, not used here
	}

	
/*****************************************************************************/
/* updatePortConfig is called when emulated app changes the serial port		**/
/* parameters baudrate, stopbits, number of databits, parity. 				**/
/*****************************************************************************/
	void CDirectSerial::updatePortConfig(Bit8u dll, Bit8u dlm, Bit8u lcr)
	{
		Bit8u parity=0;
		Bit8u bytelength=0;
		Bit16u baudrate = 0;
		
		// baud
		baudrate = dlm;
		baudrate = baudrate << 8;
		baudrate|=dll;
		if(baudrate<=0x1)			dcb.BaudRate=CBR_115200;
		else if(baudrate<=0x2)		dcb.BaudRate=CBR_57600;
		else if(baudrate<=0x3)		dcb.BaudRate=CBR_38400;
		else if(baudrate<=0x6)		dcb.BaudRate=CBR_19200;
		else if(baudrate<=0xc)		dcb.BaudRate=CBR_9600;
		else if(baudrate<=0x18)		dcb.BaudRate=CBR_4800;
		else if(baudrate<=0x30)		dcb.BaudRate=CBR_2400;
		else if(baudrate<=0x60)		dcb.BaudRate=CBR_1200;
		else if(baudrate<=0xc0)		dcb.BaudRate=CBR_600;
		else if(baudrate<=0x180)	dcb.BaudRate=CBR_300;
		else if(baudrate<=0x417)	dcb.BaudRate=CBR_110;
		
		// I read that windows can handle nonstandard baudrates:
		else dcb.BaudRate=115200/baudrate;

		// byte length
		bytelength = lcr&0x3;
		bytelength+=5;
		dcb.ByteSize=bytelength;

		// parity
		parity=lcr&0x38;
		parity = parity >> 3;
		switch(parity)
		{
		case 0x1:
			{
				dcb.Parity=ODDPARITY;
				break;
			}
		case 0x3:
			{
				dcb.Parity=EVENPARITY;
				break;
			}
		case 0x5:
			{
				dcb.Parity=MARKPARITY;
				break;
			}
		case 0x7:
			{
				dcb.Parity=SPACEPARITY;
				break;
			}
		default:
			{
				dcb.Parity=NOPARITY;
				break;
			}
		}

		// stopbits
		if(lcr&0x4)
		{
			if(bytelength==5) dcb.StopBits=ONE5STOPBITS;
			else dcb.StopBits=TWOSTOPBITS;
		}
		else
		{
			dcb.StopBits=ONESTOPBIT;
		}

		if(!SetCommState(hCom, &dcb))
			LOG_MSG("Serial port at 0x%x: API did not like the new values.");
		//LOG_MSG("Serial port at 0x%x: Port params changed: %d Baud", base,dcb.BaudRate);

	}

	void CDirectSerial::updateMSR()
	{
		Bit8u newmsr = 0;
		DWORD dptr;

		if(!GetCommModemStatus(hCom,&dptr))
		{
			//LOG_MSG("Serial port at %x: GetCommModemStatus failed!", base);
		}
		if(dptr&MS_CTS_ON) newmsr|=MSR_CTS_MASK;
		if(dptr&MS_DSR_ON) newmsr|=MSR_DSR_MASK;
		if(dptr&MS_RING_ON) newmsr|=MSR_RI_MASK;
		if(dptr&MS_RLSD_ON) newmsr|=MSR_CD_MASK;
		changeMSR(newmsr);

	}



	void CDirectSerial::transmitByte(Bit8u val)
	{
		DWORD bytesWritten;
		WriteFile(hCom, &val, 1, &bytesWritten, NULL);
		if(&bytesWritten >0)
		{
			ByteTransmitted();
			//LOG_MSG("UART 0x%x: TX 0x%x", base,val);
		}
		else
		{
			LOG_MSG("UART 0x%x: NO BYTE WRITTEN! PORT HANGS NOW!", base);
		}
	}

/*****************************************************************************/
/* setBreak(val) switches break on or off								**/
/*****************************************************************************/

	void CDirectSerial::setBreak(bool value)
	{
		//LOG_MSG("UART 0x%x: Break toggeled: %d", base, value);
		if(value) SetCommBreak(hCom);
		else ClearCommBreak(hCom);


	}

/*****************************************************************************/
/* updateModemControlLines(mcr) sets DTR and RTS.							**/
/*****************************************************************************/
	void CDirectSerial::updateModemControlLines(Bit8u mcr)
	{
		bool change = false;
		
		/*** DTR ***/
		if(mcr&MCR_DTR_MASK)	// DTR on
		{		
			if(dcb.fDtrControl==DTR_CONTROL_DISABLE)
			{
				dcb.fDtrControl=DTR_CONTROL_ENABLE;
				change=true;
			}
		}
		else
		{
			if(dcb.fDtrControl==DTR_CONTROL_ENABLE)
			{
				dcb.fDtrControl=DTR_CONTROL_DISABLE;
				change=true;
			}
		}
		/*** RTS ***/
		if(mcr&MCR_RTS_MASK)	// RTS on
		{		
			if(dcb.fRtsControl==RTS_CONTROL_DISABLE)
			{
				dcb.fRtsControl=RTS_CONTROL_ENABLE;
				change=true;
			}
		}
		else
		{
			if(dcb.fRtsControl==RTS_CONTROL_ENABLE)
			{
				dcb.fRtsControl=RTS_CONTROL_DISABLE;
				change=true;
			}
		}
		if(change)SetCommState(hCom, &dcb);
	}
	
	void CDirectSerial::Timer2(void)
	{	
		DWORD dwRead;
		DWORD errors;
		Bit8u chRead;
		


		// update RX
		if (ReadFile(hCom, &chRead, 1, &dwRead, NULL)) {
			if(dwRead != 0)
			{	
				//LOG_MSG("UART 0x%x: RX 0x%x", base,chRead);
				receiveByte(chRead);
				
			}
		}
		// check for errors
		if(ClearCommError(hCom, &errors, NULL))
		if(errors&(CE_BREAK|CE_FRAME|CE_RXPARITY))
		{
			Bit8u errreg = 0;

			LOG_MSG("Serial port: line error.");

			if(errors&CE_BREAK) errreg|= LSR_RX_BREAK_MASK;
			if(errors&CE_FRAME) errreg|= LSR_FRAMING_ERROR_MASK;
			if(errors&CE_RXPARITY) errreg|= LSR_PARITY_ERROR_MASK;

			receiveError(errreg);
		}

		// update Modem input line states
		updateMSR();

	}

		
//};

#else /*linux and others oneday maybe */

#endif 
#endif

